@@ -1,4 +1,4 @@
-(* $Id: frontend.ml 130 2009-02-16 01:09:27Z gerd $
+(* $Id: frontend.ml 159 2012-05-06 21:39:07Z gerd $
  * ----------------------------------------------------------------------
  *
  *)
@@ -292,8 +292,10 @@ let check_package_list l =
 
 let run_command ?filter verbose cmd args =
   if verbose then begin
+    let escape_if_needed s =
+      if String.contains s ' ' then "\"" ^ String.escaped s ^ "\"" else s in
     print_string ("+ " ^ cmd ^ " " ^
-		  String.concat " " args ^ "\n");
+		  String.concat " " (List.map escape_if_needed args) ^ "\n");
     if filter <> None then
       print_string ("  (output of this command is filtered by ocamlfind)\n")
   end;
@@ -314,7 +316,7 @@ let run_command ?filter verbose cmd args =
     Sys.signal Sys.sigint Sys.Signal_ignore in
 
   let need_exe =
-    List.mem Findlib_config.system [ "win32"; "mingw" ] in
+    List.mem Findlib_config.system [ "win32"; "win64"; "mingw" ] in
 
   let fixed_cmd =
     if need_exe then (
@@ -371,6 +373,46 @@ let run_command ?filter verbose cmd args =
 ;;
 
 
+(**************** Generic argument processing *************************)
+
+let merge_native_arguments native_spec f_unit f_string f_special_list =
+  List.map
+    (fun (switch_name, switch_has_arg, help_text) ->
+       let f =
+	 try
+	   List.assoc switch_name f_special_list
+	 with
+	     Not_found ->
+	       if switch_has_arg then 
+		 f_string switch_name 
+	       else 
+		 f_unit switch_name in
+       (switch_name, f, help_text)
+    )
+    native_spec
+;;
+
+
+let parse_args
+      ?(current = Arg.current) ?(args = Sys.argv) 
+      ?(align = true)
+      spec anon usage =
+  try
+    Arg.parse_argv
+      ~current
+      args
+      (if align then Arg.align spec else spec)
+      anon
+      usage
+  with
+    | Arg.Help text ->
+        print_string text;
+        exit 0
+    | Arg.Bad text ->
+        prerr_string text;
+        exit 2  
+
+
 (************************* format expansion *************************)
 
 
@@ -431,7 +473,7 @@ let query_package () =
   let i_format =
     "-I %d" in
   let l_format =
-    if Findlib_config.system = "win32" then
+    if Findlib_config.system = "win32" || Findlib_config.system = "win64" then
       (* Microsoft toolchain *)
       "-ccopt \"/link /libpath:%d\""
     else
@@ -459,17 +501,17 @@ let query_package () =
   in
 
 
-  Arg.parse
+  parse_args
     [ "-predicates", Arg.String append_predicate,
                   "      specifies comma-separated list of assumed predicates";
       "-format", Arg.String (fun s -> format := s),
-              "          specifies the output format";
+              "<fmt>      specifies the output format";
       "-separator", Arg.String (fun s -> separator := s),
                  "       specifies the string that separates multiple answers";
       "-prefix", Arg.String (fun s -> prefix := s),
-              "          a string printed before the first answer";
+              "<p>        a string printed before the first answer";
       "-suffix", Arg.String (fun s -> suffix := s),
-              "          a string printed after the last answer";
+              "<s>        a string printed after the last answer";
       "-recursive", Arg.Set recursive,
                  "       select direct and indirect ancestors/descendants, too";
       "-r", Arg.Set recursive,
@@ -582,16 +624,18 @@ let process_pp_spec syntax_preds packages pp_opts =
     if syntax_preds <> [] then
       match preprocessor_cmds with
 	  [] ->
-	    failwith("When using -syntax, the META variable 'preprocessor' must be set")
+	    failwith("Using -syntax, but no package is selected specifying \
+                     a preprocessor as required for -syntax")
 	| [_, cmd] -> Some cmd
 	| _ ->
-	    failwith("No unique value for the META variable 'preprocessor': " ^
-		     String.concat ", "
-		       (List.map
-			  (fun (n,v) ->
-			  "package " ^ n ^ " defines `" ^ v ^ "'")
-			  preprocessor_cmds
-		       )
+	    failwith("Several packages are selected that specify \
+                      preprocessors: " ^ 
+		       String.concat ", "
+		      (List.map
+			 (fun (n,v) ->
+			    "package " ^ n ^ " defines `" ^ v ^ "'")
+			 preprocessor_cmds
+		      )
 		    )
     else
       None
@@ -631,26 +675,6 @@ let process_pp_spec syntax_preds packages pp_opts =
 ;;
 
 
-(**************** Generic argument processing *************************)
-
-let merge_native_arguments native_spec f_unit f_string f_special_list =
-  List.map
-    (fun (switch_name, switch_has_arg, help_text) ->
-       let f =
-	 try
-	   List.assoc switch_name f_special_list
-	 with
-	     Not_found ->
-	       if switch_has_arg then 
-		 f_string switch_name 
-	       else 
-		 f_unit switch_name in
-       (switch_name, f, help_text)
-    )
-    native_spec
-;;
-
-
 (**************** OCAMLC/OCAMLMKTOP/OCAMLOPT subcommands ****************)
 
 type pass_file_t =
@@ -661,6 +685,13 @@ type pass_file_t =
 ;;
 
 
+let contracted_ocamlmklib_options =
+  [ "-l"; "-L"; "-R"; "-F"; "-Wl,-rpath,"; "-Wl,-R" ]
+    (* The ocamlmklib options where the argument is directly attached to the
+       switch (e.g. -L<path> instead of -L <path>)
+     *)
+
+
 let ocamlc which () =
 
   let destdir = ref (default_location()) in
@@ -701,7 +732,12 @@ let ocamlc which () =
                 pass_options := !pass_options @ [name]) in
   let add_spec_fn name s =
     pass_options := !pass_options @ [name; s] in
-  let add_spec name = Arg.String (add_spec_fn name) in
+  let add_spec name = 
+    Arg.String (add_spec_fn name) in
+  let add_contracted_spec_fn name s =
+    pass_options := !pass_options @ [name ^ s] in
+  let add_contracted_spec name =
+    Arg.String (add_contracted_spec_fn name) in
   let add_pkg =
     Arg.String (fun s -> packages := !packages @ (Fl_split.in_words s)) in
   let add_pred =
@@ -720,6 +756,7 @@ let ocamlc which () =
     match which with
       | "ocamlc"     -> Ocaml_args.ocamlc_spec
       | "ocamlcp"    -> Ocaml_args.ocamlcp_spec
+      | "ocamlmklib" -> Ocaml_args.ocamlmklib_spec
       | "ocamlmktop" -> Ocaml_args.ocamlmktop_spec
       | "ocamlopt"   -> Ocaml_args.ocamlopt_spec
       | _            -> None in
@@ -728,71 +765,108 @@ let ocamlc which () =
       | None -> failwith ("Not supported in your configuration: " ^ which)
       | Some s -> s in
 
-  Arg.parse
-    (List.flatten
-    [ [
-      "-package", add_pkg,
-               " <name>   Refer to package when compiling";
-      "-linkpkg", Arg.Set linkpkg,
-               "          Link the packages in";
-      "-predicates", add_pred,
-                  " <p>   Add predicate <p> when resolving package properties";
-      "-dontlink", add_dontlink,
-                " <name>  Do not link in package <name> and its ancestors";
-      "-syntax", add_syntax_pred,
-              " <p>       Use preprocessor with predicate <p>";
-      "-ppopt", add_pp_opt,
-             " <opt>      Append option <opt> to preprocessor invocation";
-      "-dllpath-pkg", add_dll_pkg,
-                   "<pkg> Add -dllpath for this package";
-      "-dllpath-all", Arg.Set dll_pkgs_all,
-                   "      Add -dllpath for all linked packages";
-      "-ignore-error", Arg.Set ignore_error,
-                    "     Ignore the 'error' directive in META files";
-      "-passopt", Arg.String (fun s -> pass_options := !pass_options @ [s]),
-               " <opt>    Pass option <opt> directly to ocamlc/opt/mktop\nSTANDARD OPTIONS:";
-      ];
+  let arg_spec =
+    List.flatten
+      [ [
+          "-package", add_pkg,
+            "<name>   Refer to package when compiling";
+          "-linkpkg", Arg.Set linkpkg,
+            "          Link the packages in";
+          "-predicates", add_pred,
+            "<p>   Add predicate <p> when resolving package properties";
+          "-dontlink", add_dontlink,
+            "<name>  Do not link in package <name> and its ancestors";
+          "-syntax", add_syntax_pred,
+            "<p>       Use preprocessor with predicate <p>";
+          "-ppopt", add_pp_opt,
+            "<opt>      Append option <opt> to preprocessor invocation";
+          "-dllpath-pkg", add_dll_pkg,
+            "<pkg> Add -dllpath for this package";
+          "-dllpath-all", Arg.Set dll_pkgs_all,
+            "      Add -dllpath for all linked packages";
+          "-ignore-error", Arg.Set ignore_error,
+            "     Ignore the 'error' directive in META files";
+          "-passopt", Arg.String (fun s -> pass_options := !pass_options @ [s]),
+            "<opt>    Pass option <opt> directly to ocamlc/opt/mklib/mktop";
+          "-passrest", Arg.Rest (fun s -> pass_options := !pass_options @ [s]),
+            "         Pass all remaining options directly\nSTANDARD OPTIONS:";
+        ];
 
-      merge_native_arguments 
-	native_spec 
-	add_switch
-	add_spec
-	[ "-cclib", 
-	  Arg.String (fun s -> pass_files := !pass_files @ [ Cclib s ]);
+        merge_native_arguments 
+	  native_spec 
+	  add_switch
+	  add_spec
+          (
+	    [ "-cclib", 
+	      Arg.String (fun s -> pass_files := !pass_files @ [ Cclib s ]);
+              
+	      "-I", (Arg.String
+		       (fun s ->
+		          let s = resolve_path s in
+                          if Sys.file_exists s then incpath := s :: !incpath;  (* reverted below *)
+		          add_spec_fn "-I" (slashify s) ));
+              
+	      "-impl", 
+	      Arg.String (fun s -> pass_files := !pass_files @ [ Impl(slashify s) ]);
+              
+	      "-intf", 
+	      Arg.String (fun s -> pass_files := !pass_files @ [ Intf(slashify s) ]);
+              
+	      "-pp", 
+	      Arg.String (fun s -> pp_specified := true; add_spec_fn "-pp" s);
+	      
+	      "-thread", 
+	      Arg.Unit (fun _ -> threads := threads_default);
+            
+	      "-vmthread", 
+	      Arg.Unit (fun _ -> threads := `VM_threads);
+              
+	      "-", 
+	      Arg.String (fun s -> pass_files := !pass_files @  [ Pass s ]);
+              
+	    ] @
+            if which = "ocamlmklib" then
+              List.map
+                (fun opt ->
+                   (opt, add_contracted_spec opt)
+                )
+                contracted_ocamlmklib_options
+            else
+              []
+          )
+      ] in
 
-	  "-I", (Arg.String
-		   (fun s ->
-		      let s = resolve_path s in
-                      if Sys.file_exists s then incpath := s :: !incpath;  (* reverted below *)
-		      add_spec_fn "-I" (slashify s) ));
+  let (current,args) =
+    if which = "ocamlmklib" then
+      (* Special processing for -L, -R etc. *)
+      let c = !(Arg.current) in
+      let l = Array.length Sys.argv in
+      let args1 = Array.sub Sys.argv (c+1) (l-c-1) in
+      let args2 =
+        Array.append
+          [| Sys.argv.(0) |]
+          (Fl_args.rewrite_contracted_args
+             arg_spec
+             contracted_ocamlmklib_options
+             args1
+          ) in
+      (ref 0, args2)
+    else
+      (Arg.current, Sys.argv) in
 
-	  "-impl", 
-	  Arg.String (fun s -> pass_files := !pass_files @ [ Impl(slashify s) ]);
-
-	  "-intf", 
-	  Arg.String (fun s -> pass_files := !pass_files @ [ Intf(slashify s) ]);
-
-	  "-pp", 
-	  Arg.String (fun s -> pp_specified := true; add_spec_fn "-pp" s);
-	  
-	  "-thread", 
-	  Arg.Unit (fun _ -> threads := threads_default);
-
-	  "-vmthread", 
-	  Arg.Unit (fun _ -> threads := `VM_threads);
-
-	  "-", 
-	  Arg.String (fun s -> pass_files := !pass_files @  [ Pass s ]);
-	]
-    ])
+  parse_args
+    ~current
+    ~args
+    arg_spec
     (fun s -> pass_files := !pass_files @ [ Pass s])
     ("usage: ocamlfind " ^ which ^ " [options] file ...");
 
   (* ---- Start requirements analysis ---- *)
-
+  
   begin match which with
     "ocamlc"     -> predicates := "byte" :: !predicates;
   | "ocamlcp"    -> predicates := "byte" :: !predicates;
+  | "ocamlmklib" -> predicates := "byte" :: "native" :: !predicates;
   | "ocamlmktop" -> predicates := "byte" :: "create_toploop" :: !predicates;
   | "ocamlopt"   -> predicates := "native" :: !predicates;
   | _            -> failwith "unsupported backend"
@@ -958,9 +1032,22 @@ let ocamlc which () =
 	      []
 	    else
 	      [ "-I"; slashify pkgdir;
-		"-ccopt"; out_path ~prefix:"-I" pkgdir; ])
+		(* "-ccopt"; out_path ~prefix:"-I" pkgdir; -- see comment *)
+	      ])
 	 eff_packages_dl) in
+  (* We no longer emit -ccopt options, because ocamlc/ocamlopt already
+     do that for each -I if the C compiler needs to be invoked
+     (so far I tracked it, ocamlc/ocamlopt have always done this, even
+      back in 1996).
+   *)
 
+  let l_options = [] in
+  (* Also, no longer -ccopt -L options. Current ocamlc/ocamlopt do that
+     for each -I option passed to them anyway, so we can omit that here.
+     See ocaml change (quite old, but I was not aware of it):
+      http://camlcvs.inria.fr/cgi-bin/cvsweb/ocaml/asmcomp/asmlink.ml.diff?r1=1.38;r2=1.39
+   *)
+(*
   let l_options =
     List.flatten
       (List.map
@@ -969,12 +1056,13 @@ let ocamlc which () =
 	    if List.mem npkgdir exclude_list then
 	      []
 	    else
-	      if Findlib_config.system = "win32" then
+	      if Findlib_config.system = "win32" || Findlib_config.system = "win64" then
 		(* Microsoft toolchain *)
 		[ "-ccopt"; out_path ~prefix:"/link /libpath:" pkgdir ]
 	      else
 		[ "-ccopt"; out_path ~prefix:"-L" pkgdir; ])
 	 eff_link_dl) in
+ *)
 
   let archives =
     List.flatten
@@ -1050,7 +1138,12 @@ let ocamlc which () =
 	 (fun pkg -> ["-dllpath";  slashify pkg] )
 	 dll_dirs) in
 
+  let mklib_options =
+    ["-ocamlc"; Findlib.command `ocamlc;
+     "-ocamlopt"; Findlib.command `ocamlopt] in
+
   let arguments =
+    (if which = "ocamlmklib" then mklib_options else []) @
     !pass_options @    (* other options from the command line *)
     i_options @        (* Generated -I options from package analysis *)
     pp_command @       (* Optional preprocessor command *)
@@ -1066,6 +1159,7 @@ let ocamlc which () =
 	"ocamlc"     -> Findlib.command `ocamlc
       | "ocamlopt"   -> Findlib.command `ocamlopt
       | "ocamlcp"    -> Findlib.command `ocamlcp
+      | "ocamlmklib" -> Findlib.command `ocamlmklib
       | "ocamlmktop" -> Findlib.command `ocamlmktop
       | _            -> assert false
   in
@@ -1093,30 +1187,48 @@ let ocamldoc() =
       | None -> failwith "Not supported in your configuration: ocamldoc"
       | Some s -> s in
 
-  Arg.parse
-    ( [ "-package",
-	Arg.String (fun s -> 
-		      packages := Fl_split.in_words s @ !packages),
-	"<name>  Add this package to the search path";
+  parse_args
+    ~align:false
+    ( Arg.align
+        [ "-package",
+	  Arg.String (fun s -> 
+		        packages := Fl_split.in_words s @ !packages),
+	  "<name>  Add this package to the search path";
+          
+	  "-predicates",
+	  Arg.String (fun s ->
+		        predicates := Fl_split.in_words s @ !predicates),
+	  "<p>  Add predicate <p> when calculating dependencies";
 
-	"-predicates",
-	Arg.String (fun s ->
-		      predicates := Fl_split.in_words s @ !predicates),
-	"<p>  Add predicate <p> when calculating dependencies";
+	  "-syntax",
+	  Arg.String (fun s ->
+		        syntax_preds := Fl_split.in_words s @ !syntax_preds),
+	  "<p>  Use preprocessor with predicate <p>";
 
-	"-syntax",
-	Arg.String (fun s ->
-		      syntax_preds := Fl_split.in_words s @ !syntax_preds),
-	"<p>      Use preprocessor with predicate <p>";
-
-	"-ppopt",
-	Arg.String (fun s -> pp_opts := s :: !pp_opts),
-	"<opt>     Append option <opt> to preprocessor invocation";
-
-	"-verbose",
-	Arg.Set verbose,
-	"        Be verbose\nSTANDARD OPTIONS:";
-      ]
+	  "-ppopt",
+	  Arg.String (fun s -> pp_opts := s :: !pp_opts),
+	  "<opt>  Append option <opt> to preprocessor invocation";
+          
+	  "-thread",
+	  Arg.Unit (fun () -> predicates := "mt" :: "mt_posix" :: !predicates),
+	  "  Assume kernel multi-threading when doing dependency analyses";
+          
+	  "-vmthread",
+	  Arg.Unit (fun () -> predicates := "mt" :: "mt_vm" :: !predicates),
+	  "  Assume bytecode multi-threading when doing dependency analyses";
+          
+          "-passopt",
+          Arg.String (fun s -> options := !options @ [s]),
+          "<opt>  Pass this option directly to ocamldoc";
+          
+          "-passrest",
+          Arg.Rest (fun s -> options := !options @ [s]),
+          "  Pass all remaining options directly to ocamldoc";
+          
+	  "-verbose",
+	  Arg.Set verbose,
+	  "  Be verbose\nSTANDARD OPTIONS:";
+        ]
       @
       ( merge_native_arguments
 	  native_spec
@@ -1164,6 +1276,18 @@ let ocamldoc() =
   let eff_packages =
     package_deep_ancestors !predicates !packages in
 
+  (* Check on [error] directives (turned into warnings): *)
+  List.iter
+    (fun pkg ->
+       try
+	 let error = package_property !predicates pkg "error" in
+	 prerr_endline("ocamlfind: [WARNING] Package `" ^ pkg ^
+			 "' signals error: " ^ error)
+       with
+	   Not_found -> ()
+    )
+    eff_packages;
+
   let eff_packages_dl =
     remove_dups (List.map package_directory eff_packages) in
 
@@ -1252,16 +1376,18 @@ let ocamldep () =
       | None -> failwith "Not supported in your configuration: ocamldep"
       | Some s -> s in
 
-  Arg.parse
+  parse_args
     ( [
 	"-syntax", add_syntax_pred,
-                " <p>       Use preprocessor with predicate <p>";
+                "<p>       Use preprocessor with predicate <p>";
 	"-package", add_pkg,
-	         " <p>      Add preprocessor package <p>";
+	         "<p>      Add preprocessor package <p>";
 	"-ppopt", add_pp_opt,
-               " <opt>      Append option <opt> to preprocessor invocation";
+               "<opt>      Append option <opt> to preprocessor invocation";
 	"-passopt", Arg.String (fun s -> pass_options := !pass_options @ [s]),
-                 " <opt>    Pass option <opt> directly to ocamlc/opt/mktop";
+                 "<opt>    Pass option <opt> directly to ocamlc/opt/mktop";
+        "-passrest", Arg.Rest (fun s -> pass_options := !pass_options @ [s]),
+                  "         Pass all remaining options directly";
 	"-native-filter", Arg.Set native_filter,
 	               "    Output only dependencies for native code (implies -native)";
 	"-bytecode-filter", Arg.Set bytecode_filter,
@@ -1352,16 +1478,18 @@ let ocamlbrowser () =
   let add_pkg =
     Arg.String (fun s -> packages := !packages @ (Fl_split.in_words s)) in
 
-  Arg.parse
+  parse_args
       [
 	"-I", Arg.String (fun s -> add_spec_fn "-I" (slashify(resolve_path s))),
-           " <dir>          Add <dir> to the list of include directories";
+           "<dir>          Add <dir> to the list of include directories";
 	"-all", Arg.Set add_all,
 	     "              Add all packages to include path";
 	"-package", add_pkg,
-	         " <p>      Add package <p> to include path";
+	         "<p>      Add package <p> to include path";
 	"-passopt", Arg.String (fun s -> pass_options := !pass_options @ [s]),
-                 " <opt>    Pass option <opt> directly to ocamlbrowser";
+                 "<opt>    Pass option <opt> directly to ocamlbrowser";
+        "-passrest", Arg.Rest (fun s -> pass_options := !pass_options @ [s]),
+                 "          Pass all remaining options directly";
       ]
       (fun s -> raise (Arg.Bad ("Unexpected argument: " ^ s)))
       ("usage: ocamlfind ocamlbrowser [options] file ...");
@@ -1621,7 +1749,7 @@ let install_package () =
     ] in
   let errmsg = "usage: ocamlfind install [options] <package_name> <file> ..." in
 
-  Arg.parse
+  parse_args
         keywords
 	(fun s ->
 	   if !pkgname = ""
@@ -1832,7 +1960,7 @@ let remove_package () =
     ] in
   let errmsg = "usage: ocamlfind remove [options] <package_name>" in
 
-  Arg.parse
+  parse_args
         keywords
 	(fun s ->
 	   if !pkgname = ""
@@ -1936,7 +2064,7 @@ let list_packages() =
     ] in
   let errmsg = "usage: ocamlfind list [options]" in
 
-  Arg.parse
+  parse_args
       keywords
       (fun _ -> Arg.usage keywords errmsg; exit 1)
       errmsg;
@@ -1957,7 +2085,7 @@ let print_configuration() =
   let var = ref None in
   let errmsg = "usage: ocamlfind printconf (conf|path|destdir|metadir|stdlib|ldconf)" in
 
-  Arg.parse
+  parse_args
         []
 	(fun s ->
 	   if !var <> None then raise(Arg.Bad "Unexpected argument");
@@ -2035,6 +2163,7 @@ let rec select_mode () =
     | ("remove"|"-remove")                 -> incr Arg.current; M_remove
     | ("ocamlc"|"-ocamlc"|"c")             -> incr Arg.current; M_compiler "ocamlc"
     | ("ocamlcp"|"-ocamlcp"|"cp")          -> incr Arg.current; M_compiler "ocamlcp"
+    | ("ocamlmklib"|"-ocamlmklib"|"mklib") -> incr Arg.current; M_compiler "ocamlmklib"
     | ("ocamlmktop"|"-ocamlmktop"|"mktop") -> incr Arg.current; M_compiler "ocamlmktop"
     | ("ocamlopt"|"-ocamlopt"|"opt")       -> incr Arg.current; M_compiler "ocamlopt"
     | ("ocamldep"|"-ocamldep"|"dep")       -> incr Arg.current; M_dep
@@ -2080,16 +2209,17 @@ let main() =
     | M_remove         -> remove_package ()
     | M_printconf      -> print_configuration ()
     | M_list           -> list_packages()
-    | M_compiler which -> ocamlc which ()
     | M_dep            -> ocamldep()
     | M_browser        -> ocamlbrowser()
     | M_doc            -> ocamldoc()
     | M_call(pkg,cmd)  -> ocamlcall pkg cmd
+    | M_compiler which -> ocamlc which ()
   with
     Usage ->
       prerr_endline "Usage: ocamlfind query        [-help | other options] <package_name> ...";
       prerr_endline "   or: ocamlfind ocamlc       [-help | other options] <file> ...";
       prerr_endline "   or: ocamlfind ocamlcp      [-help | other options] <file> ...";
+      prerr_endline "   or: ocamlfind ocamlmklib   [-help | other options] <file> ...";
       prerr_endline "   or: ocamlfind ocamlmktop   [-help | other options] <file> ...";
       prerr_endline "   or: ocamlfind ocamlopt     [-help | other options] <file> ...";
       prerr_endline "   or: ocamlfind ocamldep     [-help | other options] <file> ...";
